.md-header__button.md-logo {
    margin: 0px;
    padding: 0;
}

.md-header__button.md-logo img, .md-header__button.md-logo img {
    height: 2.4rem;
    width: 2.4rem;
}

#on_being_together {
    font-family: inherit;
}